# Copyright 2023-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Isabelle Connector
===================

A connector to the Isabelle server, hiding server interactions.
"""  # noqa: D205, D400

import json
import logging
from collections.abc import Sequence
from typing import Optional
from uuid import uuid4

from isabelle_client.socket_communication import IsabelleResponseType
from isabelle_client.utils import (
    get_isabelle_client,
    get_or_create_working_directory,
    start_isabelle_server,
)


class IsabelleTheoryError(RuntimeError):
    """Raised when the Isabelle response contains errors."""


class IsabelleConnector:
    r"""
    A connector to the Isabelle server, hiding server interactions.

    :param working_directory: a directory for storing the server logs,
            temporary theory files etc.

    >>> import os
    >>> os.environ["PATH"] = "isabelle_client/resources:$PATH"
    >>> connector = IsabelleConnector()
    >>> print(connector.working_directory)
    /...
    >>> connector.build_theory(
    ...     'lemma "\<forall> x. \<exists> y. x = y" by auto', theory="Mock")
    >>> connector.build_theory(
    ...     'lemma "\<forall> x. \<forall> y. x = y" by auto', theory="Fail")
    Traceback (most recent call last):
     ...
    isabelle...Error: Failed to finish proof\<^here>:
    goal (1 subgoal):
     1. \<And>x y. x = y
    """

    def __init__(self, working_directory: Optional[str] = None) -> None:  # noqa: D107
        self._working_directory = get_or_create_working_directory(
            working_directory
        )
        server_info, self._server_process = start_isabelle_server(
            log_file=str(self._working_directory / "isabelle-server.log")
        )
        self._client = get_isabelle_client(server_info=server_info)
        self._client.logger = logging.getLogger()
        self._client.logger.setLevel(logging.INFO)
        self._client.logger.addHandler(
            logging.FileHandler(str(self._working_directory / "session.log"))
        )

    def _write_temp_theory_file(
        self,
        theory_body: str,
        imports: Sequence[str],
        theory: Optional[str] = None,
    ) -> str:
        theory_name = (
            "T" + str(uuid4()).replace("-", "") if theory is None else theory
        )
        (self._working_directory / f"{theory_name}.thy").write_text(
            f"theory {theory_name}\n"
            f"imports {', '.join(imports)}\n"
            f"begin\n{theory_body}\nend\n"
        )
        return theory_name

    def build_theory(
        self,
        theory_body: str,
        imports: Sequence[str] = ("Main",),
        theory: Optional[str] = None,
    ) -> None:
        """
        Build a theory using the Isabelle server.

        :param theory_body: theory body (goes between begin and end keywords)
        :param imports: which theories to import
        :param theory: (for tests) fixed named for theory file
        :raises IsabelleTheoryError: if validation failed
        """
        theory_name = self._write_temp_theory_file(
            theory_body=theory_body,
            imports=imports,
            theory=theory,
        )
        validation_result = self._client.use_theories(
            theories=[theory_name], master_dir=str(self._working_directory)
        )
        for isabelle_response in validation_result:
            if (
                isabelle_response.response_type
                == IsabelleResponseType.FINISHED
            ):
                json_response = json.loads(isabelle_response.response_body)
                if errors := json_response["errors"]:
                    raise IsabelleTheoryError(errors[0]["message"])

    @property
    def working_directory(self) -> str:
        """Get working directory."""
        return str(self._working_directory)



# Copyright 2021-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities
==========

A collection of different useful functions.
"""  # noqa: D205, D400

import asyncio
import json
import re
import socketserver
import sys
import tempfile
from enum import Enum
from importlib.resources import files
from pathlib import Path
from typing import Optional
from uuid import uuid4

from isabelle_client.isabelle__client import IsabelleClient

MS_WINDOWS = "win32"


class IsabelleServerCommands(Enum):
    """Supported Isabelle server commands."""

    HELP = "help"
    USE_THEORIES = "use_theories"
    ECHO = "echo"


def get_isabelle_client(server_info: str) -> IsabelleClient:
    """
    Get an instance of ``IsabelleClient`` from server info.

    >>> server_inf = 'server "test" = 127.0.0.1:10000 (password "pass")'
    >>> print(get_isabelle_client(server_inf).port)
    10000
    >>> get_isabelle_client("wrong")
    Traceback (most recent call last):
        ...
    ValueError: Unexpected server info: wrong

    :param server_info: a line returned by a server on start
    :returns: an Isabelle client
    :raises ValueError: if the server response is malformed
    """
    match = re.compile(
        r"server \".*\" = (.*):(.*) \(password \"(.*)\"\)"
    ).match(server_info)
    if match is None:
        msg = f"Unexpected server info: {server_info}"
        raise ValueError(msg)
    address, port, password = match.groups()
    return IsabelleClient(address, int(port), password)


def start_isabelle_server(
    log_file: Optional[str] = None,
    name: Optional[str] = None,
    port: Optional[int] = None,
) -> tuple[str, asyncio.subprocess.Process]:
    """
    Start Isabelle server.

    >>> import os
    >>> os.environ["PATH"] = "isabelle_client/resources:$PATH"
    >>> print(start_isabelle_server()[0])
    server "isabelle" = 127.0.0.1:9999 (password "test_password")
    <BLANKLINE>

    :param log_file: a log file for exceptional output of internal server and
        session operations
    :param name: explicit server name (default: isabelle)
    :param port: explicit server port
    :returns: a line of server info and server process
    """
    args = (
        "server"
        + (f" -L {log_file}" if log_file is not None else "")
        + (f" -p {port!s}" if port is not None else "")
        + (f" -n {name}" if name is not None else "")
    )
    if sys.platform == MS_WINDOWS:
        return start_isabelle_server_win32(args)  # pragma: no cover

    async def async_call() -> tuple[str, asyncio.subprocess.Process]:
        """
        Start Isabelle server asynchronously.

        :returns: a line of server info and server process
        :raises ValueError: if no stdout seen after starting the server
        """
        isabelle_server = await asyncio.create_subprocess_exec(
            "isabelle", *(args.split(" ")), stdout=asyncio.subprocess.PIPE
        )
        if isabelle_server.stdout is not None:
            return (await isabelle_server.stdout.readline()).decode(
                "utf-8"
            ), isabelle_server
        msg = "No stdout while starting the server."  # pragma: no cover
        raise ValueError(msg)  # pragma: no cover

    return asyncio.run(async_call())


def start_isabelle_server_win32(
    args: str,
) -> tuple[str, asyncio.subprocess.Process]:  # pragma: no cover
    """
    Start Isabelle server on Windows.

    :param args: Isabelle server arguments string
    :returns: a line of server info and server process
    """
    # this line enables asyncio.create_subprocess_exec on Windows:
    # https://docs.python.org/3/library/asyncio-platforms.html#asyncio-windows-subprocess
    asyncio.set_event_loop_policy(
        asyncio.WindowsProactorEventLoopPolicy()  # type: ignore
    )

    async def async_call() -> tuple[str, asyncio.subprocess.Process]:
        """
        Start Isabelle server in Cygwin asynchronously.

        :returns: a line of server info and server process
        :raises ValueError: if no stdout seen after starting the server
        """
        isabelle_server = await asyncio.create_subprocess_exec(
            str(
                files("isabelle_client")
                .joinpath("resources")
                .joinpath("Cygwin-Isabelle.bat")
            ),
            args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        if isabelle_server.stdout is not None:
            server_info = (await isabelle_server.stdout.readline()).decode(
                "utf-8"
            )
            return server_info, isabelle_server
        msg = "No stdout while starting the server."
        raise ValueError(msg)

    return asyncio.run(async_call())


class BuggyDummyTCPHandler(socketserver.BaseRequestHandler):
    """A dummy handler to mock bugs in Isabelle server response."""

    def handle(self) -> None:
        """Return something weird."""
        request = self.request.recv(4096).decode("utf-8").split("\n")[1]
        if request == IsabelleServerCommands.HELP.value:
            self.request.sendall(b"5\n")
            self.request.sendall(b"# !!!")
        else:
            self.request.sendall(
                b'OK {"isabelle_id":"mock","isabelle_name":"Isabelle2024"}\n'
            )
            self.request.sendall(b"ERROR UNEXPECTED\n")


class DummyTCPHandler(socketserver.BaseRequestHandler):
    """A dummy handler to mock Isabelle server."""

    def _mock_command_execution(self, command: str, arguments: str) -> None:
        filename = command
        if (
            command == IsabelleServerCommands.USE_THEORIES.value
            and (theory_name := json.loads(arguments)["theories"][0]) != "Mock"
        ):
            filename += f".{theory_name}"  # type: ignore
        self.request.sendall(
            files("isabelle_client")
            .joinpath("resources")
            .joinpath("isabelle-responses")
            .joinpath(filename)
            .read_text()
            .encode()
        )

    def handle(self) -> None:
        """Return something similar to what Isabelle server does."""
        request = self.request.recv(4096).decode("utf-8").split("\n")[1]
        command = request.split(" ")[0]
        arguments = request[len(command) :]
        self.request.sendall(
            b'OK {"isabelle_id":"mock","isabelle_name":"Isabelle2024"}\n'
        )
        if command == IsabelleServerCommands.ECHO.value:
            self.request.sendall(f"OK {arguments[1:]}\n".encode())
        else:
            self._mock_command_execution(command, arguments)


class ReusableDummyTCPServer(socketserver.TCPServer):
    """Ignore TIME-WAIT during testing."""

    allow_reuse_address = True


def get_or_create_working_directory(working_directory: Optional[str]) -> Path:
    """
    Get existing or create a randomly named directory.

    :param working_directory: directory name
    :returns: (possibly new) directory name
    """
    new_working_directory = (
        Path(working_directory)
        if working_directory is not None
        else Path(tempfile.mkdtemp()) / str(uuid4())
    )
    if not Path(new_working_directory).exists():
        Path(new_working_directory).mkdir()
    return new_working_directory



# Copyright 2021-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fixtures for unit tests live here."""

import threading
from collections.abc import Generator
from unittest.mock import Mock

import pytest

from isabelle_client.utils import (
    BuggyDummyTCPHandler,
    DummyTCPHandler,
    ReusableDummyTCPServer,
)


@pytest.fixture(autouse=True, scope="session")
def tcp_servers() -> Generator[
    tuple[ReusableDummyTCPServer, ReusableDummyTCPServer], None, None
]:
    """
    Get a simplistic TCP server mocking Isabelle server behaviour.

    :yields: an instance of a mock working server and a mock buggy server
    """
    with (
        ReusableDummyTCPServer(("localhost", 9999), DummyTCPHandler) as server,
        ReusableDummyTCPServer(
            ("localhost", 9998), BuggyDummyTCPHandler
        ) as buggy_server,
    ):
        thread = threading.Thread(target=server.serve_forever)
        thread.daemon = True
        thread.start()
        buggy_thread = threading.Thread(target=buggy_server.serve_forever)
        buggy_thread.daemon = True
        buggy_thread.start()
        yield server, buggy_server


@pytest.fixture
def mock_logger() -> Mock:
    """
    Get a mock for logger to spy on ``info`` calls.

    :returns: mock logger
    """
    logger = Mock()
    logger.info = Mock()
    return logger



# Copyright 2021-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Socket Communication
=====================

A collection of functions for TCP communication.
"""  # noqa: D205, D400

import asyncio
import re
from collections.abc import AsyncGenerator
from dataclasses import dataclass
from enum import Enum
from logging import Logger
from typing import Optional


class IsabelleResponseType(Enum):
    """Isabelle server response type."""

    OK = "OK"
    FINISHED = "FINISHED"
    NOTE = "NOTE"
    FAILED = "FAILED"
    ERROR = "ERROR"


ASYNCHRONOUS_FINAL_MESSAGES = {
    IsabelleResponseType.FAILED,
    IsabelleResponseType.FINISHED,
    IsabelleResponseType.ERROR,
}
SYNCHRONOUS_FINAL_MESSAGES = {
    IsabelleResponseType.OK,
    IsabelleResponseType.ERROR,
}


@dataclass
class IsabelleResponse:
    """
    A response from an Isabelle server.

    .. attribute :: response_type

        an all capitals word like ``FINISHED`` or ``ERROR``

    .. attribute :: response_body

         a JSON-formatted response

    .. attribute :: response_length

        a length of JSON response
    """

    response_type: IsabelleResponseType
    response_body: str
    response_length: Optional[int] = None

    def __str__(self) -> str:
        """
        Pretty print Isabelle server response.

        :returns: a string representation of Isabelle server response
        """
        return (
            (
                f"{self.response_length}\n"
                if self.response_length is not None
                else ""
            )
            + self.response_type.value
            + (" " if self.response_body else "")
            + self.response_body
        )


async def get_response_from_isabelle(
    reader: asyncio.StreamReader,
) -> IsabelleResponse:
    r"""
    Get a response from Isabelle server.

    Format:

    * a carriage-return delimited message or
    * a fixed length message after a carriage-return delimited message with
      only one integer number denoting length

    >>> async def awaiter():
    ...     test_reader, test_writer = await asyncio.open_connection(
    ...     "localhost", 9999
    ... )
    ...     test_writer.write(b"test_password\nhelp\n")
    ...     result = [str(await get_response_from_isabelle(test_reader))]
    ...     result += [str(await get_response_from_isabelle(test_reader))]
    ...     return result
    >>> print(asyncio.run(awaiter()))
    ['OK {"isabelle_id":"mock","isabelle_name":"Isabelle2024"}', '118\nOK [...]
    >>> async def awaiter():
    ...     test_reader, test_writer = await asyncio.open_connection(
    ...     "localhost", 9998
    ... )
    ...     test_writer.write(b"test_password\nhelp\n")
    ...     result = [str(await get_response_from_isabelle(test_reader))]
    ...     result += [str(await get_response_from_isabelle(test_reader))]
    ...     return result
    >>> print(asyncio.run(awaiter()))
    Traceback (most recent call last):
      ...
    ValueError: Unexpected response from Isabelle: # !!!

    :param reader: a StreamReader connected to a server
    :returns: a response from Isabelle
    :raises ValueError: if the server response is malformed
    """
    response = (await reader.readline()).decode("utf-8")
    match = re.compile(r"(\d+)\n").match(response)
    length = int(match.group(1)) if match is not None else None
    if length is not None:
        response = (await reader.readexactly(length)).decode("utf-8")
    if (match := re.compile(r"(\w+) ?(.*)").match(response)) is None:
        msg = f"Unexpected response from Isabelle: {response}"
        raise ValueError(msg)
    return IsabelleResponse(
        IsabelleResponseType(match.group(1)), match.group(2), length
    )


async def get_final_message(
    reader: asyncio.StreamReader,
    final_message: set[IsabelleResponseType],
    logger: Optional[Logger] = None,
) -> AsyncGenerator[IsabelleResponse]:
    r"""
    Get responses from Isabelle server.

    (until a message of specified 'final' type arrives)

    >>> test_logger = getfixture("mock_logger")
    >>> async def awaiter():
    ...     test_reader, test_writer = await asyncio.open_connection(
    ...     "localhost", 9999
    ... )
    ...     test_writer.write(b"test_password\nhelp\n")
    ...     result = []
    ...     async for message in get_final_message(
    ...         test_reader, {IsabelleResponseType.OK}, test_logger
    ...     ):
    ...         result.append(message)
    ...     return result
    >>> for response in asyncio.run(awaiter()):
    ...     print(response)
    OK {"isabelle_id":"mock","isabelle_name":"Isabelle2024"}
    118
    OK ["cancel","echo","help","purge_theories","session_build",...]
    >>> print(test_logger.info.mock_calls)
    [call('OK {"isabelle_id":"mock","isabelle_name":"Isabelle2024"}'),
     call('118\nOK ["cancel","echo","help","purge_theories","session_buil...')]

    :param reader: a ``StreamReader`` connected to Isabelle server
    :param final_message: a set of possible final message types
    :param logger: a logger where to send all server replies
    :yields: the final response from Isabelle server
    """
    response = IsabelleResponse(IsabelleResponseType.NOTE, "")
    password_ok_received = False
    while (
        response.response_type not in final_message or not password_ok_received
    ):
        if response.response_type == IsabelleResponseType.OK:
            password_ok_received = True
        response = await get_response_from_isabelle(reader)
        if logger is not None:
            logger.info(str(response))
        yield response
